from mechanize import Browser
from lxml import etree
import sys
import cookielib
import os
import xml.etree.ElementTree
import psycopg2

TEST_FILE_DEF = "test.txt"
DB_FILE_DEF = "market_ddl.sql"

def db_connect(conn_string):
    try:
        conn = psycopg2.connect(conn_string)
        conn.autocommit = True
        db = conn.cursor()
    except Exception, inst:
        raise Exception("Internal error: failed to connect to db with params " + conn_string + " because of " + str(inst))
    return (conn, db)

def exec_and_retrieve(db, sql, output = True):
    result = []
    try:
        db.execute(sql)
        if (output):
            res = db.fetchone()
            while (res != None):
                result += [{}]
                for j in range(len(db.description)):
                    if (res[j] == None):
                        result[-1][db.description[j].name] = None
                    else:
                        result[-1][db.description[j].name] = res[j]
                res = db.fetchone()
    except Exception, inst:
        raise Exception("Internal error while executing SQL command\n\t" + sql + "\nThe error is:\n\t" + str(inst))
    return result

def check_syntax(html, parser, lineno, save_anyway = None):
    try:
        tree = etree.XML(html, parser)
        if save_anyway:
            f = open(save_anyway, "w")
            f.write(html)
            f.close()
        return True
    except Exception as inst:
        print "HTML file obtained in line %d does not pass HTML validation, saved in file errors/error.%d.html" %(lineno, lineno)
        f = open("errors/error.%d.html" %lineno, "w")
        f.write(html)
        f.close()
        print "Error explanation: " + str(inst)
        return False

def xml_modify_rec(root, tag, value):
    if len(root) == 2:
        for i in [0, 1]:
            if root[i].tag == "param-name" and root[i].text == tag:
                root[(i + 1)%2].text = value
                return
    for child in root:
        xml_modify_rec(child, tag, value)

def xml_modify(xml_file, tag, value):
    root = xml.etree.ElementTree.parse(xml_file).getroot()
    xml_modify_rec(root, tag, value)
    content = xml.etree.ElementTree.tostring(root)
    f = open(xml_file, "w")
    f.write(content)
    f.close()

def get_image(html, line_ar):
    src_image = None
    count = int(line_ar[1])
    while count > 0 and src_image != -1:
        src_image = html[(0 if src_image == None else src_image + 1):].find("<img") + (0 if src_image == None else src_image + 1)
        count -= 1
    if src_image == -1 or src_image == None:
        return None
    close_tag = html[src_image:].find("/>")
    src_attr = html[src_image + 1:].find("src=")
    if close_tag < src_attr:
        return None
    src_image = (html[src_image:].split('src="'), html[src_image:].split("src='"))
    if len(src_image[0]) == 1 and len(src_image[1]) == 1:
        return None
    elif len(src_image[0]) != 1:
        return src_image[0][1].split('"')[0]
    else:
        return src_image[1][1].split("'")[0]

def get_struct(root):
    struct = {}
    struct["accounts"] = []
    struct["categories"] = []
    struct["products"] = []
    for account in root[0]:
        struct["accounts"] += [account.attrib]
    for cat in root[1]:
        struct["categories"] += [{"name": cat.attrib["name"]}]
        for elem in cat:
            if elem.tag == "{http://melatti.org/market}id":
                struct["categories"][-1]["id"] = elem.text
            elif elem.tag == "{http://melatti.org/market}image":
                if "src" in elem.attrib:
                    struct["categories"][-1]["image_src"] = elem.attrib["src"]
                if "alt" in elem.attrib:
                    struct["categories"][-1]["image_alt"] = elem.attrib["alt"]
            elif elem.tag == "{http://melatti.org/market}item":
                struct["products"] += [{"show": "show" not in elem.attrib or elem.attrib["show"] == "true", "idCat": struct["categories"][-1]["id"]}]
                for item_elem in elem:
                    if item_elem.tag != "{http://melatti.org/market}currency" and item_elem.tag != "{http://melatti.org/market}image":
                        struct["products"][-1][item_elem.tag.replace("{http://melatti.org/market}", "")] = item_elem.text
                    elif item_elem.tag == "{http://melatti.org/market}currency":
                        for curr in item_elem:
                            struct["products"][-1]["currency_" + curr.tag.replace("{http://melatti.org/market}", "")] = curr.text
                    else:
                        if "src" in item_elem.attrib:
                            struct["products"][-1]["image_src"] = item_elem.attrib["src"]
                        if "alt" in item_elem.attrib:
                            struct["products"][-1]["image_alt"] = item_elem.attrib["alt"]
    return struct

def from_xml_to_sql(db, xml_file):
    root = xml.etree.ElementTree.parse(xml_file).getroot()
    struct = get_struct(root)
    sql = ""
    for account in struct["accounts"]:
        sql += "('%s', '%s'" %(account["username"], account["password"])
        if "realname" in account:
            sql += ", '%s'" %(account["realname"])
        else:
            sql += ", NULL"
        if "realsurname" in account:
            sql += ", '%s'" %(account["realsurname"])
        else:
            sql += ", NULL"
        sql += ", %s), " %(str(account["accountType"] == "administrator"))
    if sql:
        exec_and_retrieve(db, "INSERT INTO account (username, password, realname, realsurname, is_admin) VALUES " + sql[:-2], False)

    sql = ""
    cat_ids = []
    for cat in struct["categories"]:
        sql += "(%s, '%s'" %(cat["id"], cat["name"])
        cat_ids += [cat["id"]]
        if "image_src" in cat:
            sql += ", '%s'" %(cat["image_src"])
        else:
            sql += ", NULL"
        if "image_alt" in cat:
            sql += ", '%s'" %(cat["image_alt"])
        else:
            sql += ", NULL"
        sql += "), "
    if sql:
        exec_and_retrieve(db, "INSERT INTO category (id, categ_name, image_path, image_descr) VALUES " + sql[:-2], False)

    sql = ""
    sql2 = ""
    currencies = {}
    item_id = 1
    for item in struct["products"]:
        sql += "(%d, '%s', %s, '%s', TIMESTAMP WITHOUT TIME ZONE '%s', %s" %(item_id, item["title"], item["price"], item["currency_name"], item["available"], str(item["show"]))
        sql2 += "(%d, %s), " %(item_id, item["idCat"])
        item_id += 1
        if item["currency_name"] not in currencies:
            currencies[item["currency_name"]] = {}
            if "currency_toEUR" in item:
                currencies[item["currency_name"]]["toEUR"] = item["currency_toEUR"]
            if "currency_toDollar" in item:
                currencies[item["currency_name"]]["toDollar"] = item["currency_toDollar"]
        if "description" in item:
            sql += ", '%s'" %(item["description"])
        else:
            sql += ", NULL"
        if "image_src" in item:
            sql += ", '%s'" %(item["image_src"])
        else:
            sql += ", NULL"
        if "image_alt" in item:
            sql += ", '%s'" %(item["image_alt"])
        else:
            sql += ", NULL"
        sql += "), "
    if sql:
        sql_curr = ""
        for curr in currencies:
            sql_curr += "('%s'" %(curr)
            if "toEUR" in currencies[curr]:
                sql_curr += ", %s" %(currencies[curr]["toEUR"])
            else:
                sql_curr += ", NULL"
            if "toDollar" in currencies[curr]:
                sql_curr += ", %s" %(currencies[curr]["toDollar"])
            else:
                sql_curr += ", NULL"
            sql_curr += "), "
        if sql_curr:
            exec_and_retrieve(db, "INSERT INTO currency (curr_name, to_EUR, to_dollar) VALUES " + sql_curr[:-2], False)
        exec_and_retrieve(db, "INSERT INTO item (id, title, price, currency, available, show, description, image_path, image_descr) VALUES " + sql[:-2], False)
        exec_and_retrieve(db, "INSERT INTO categ_item (item_id, categ_id) VALUES " + sql2[:-2], False)

if len(sys.argv) == 3:
    test_file = sys.argv[1]
    db_file = sys.argv[2]
else:
    test_file = TEST_FILE_DEF
    db_file = DB_FILE_DEF
try:
    f = open("../WEB-INF/source/Makefile", "r")
    f.close()
except Exception as ex:
    print 'Unable to find file ../WEB-INF/source/Makefile, please refer to Section "Modalita\' di Consegna" of the project specifications'
    sys.exit(1)
try:
    f = open(test_file, "r")
except Exception as ex:
    print "Unable to open file test.txt because: " + str(ex)
    sys.exit(1)
test = f.readlines()
f.close()
(conn, db) = db_connect("dbname=ecommerce_2016 host=localhost user=ecommerce_2016_user password=ecommerce_2016_user_password port=5432")
parser = etree.XMLParser(ns_clean=True)
br = Browser()
cj = cookielib.CookieJar()
br.set_cookiejar(cj)
#br.addheaders = [('User-agent', 'Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.1) Gecko/2008071615 Fedora/3.0.1-1.fc9 Firefox/3.0.1')]
lineno = 0
curr_test_passed = True
curr_test = 1
total_points = 0
points = 0
for line in test:
    try:
        lineno += 1
        if line[0] == '#':
            continue
        line = line.replace("\n", "")
        line_ar = line.split("|")
        if line_ar[0] == "test_passed":
            if curr_test_passed:
                points += int(line_ar[1])
                print "Test " + str(int(curr_test)) + " passed! Added " + str(int(line_ar[1])) + " points, current score is " + str(int(points)) + "/" + str(total_points + int(line_ar[1]))
            else:
                print "Test " + str(int(curr_test)) + " did not pass! Lost " + str(int(line_ar[1])) + " points, current score is " + str(int(points)) + "/" + str(total_points + int(line_ar[1]))
            total_points += int(line_ar[1])
            curr_test += 1
            curr_test_passed = True
        elif curr_test_passed:
            if line_ar[0] == "copy":
                filename = os.path.join(*(line_ar[1].split("/")))
                try:
                    f = open(filename, "r")
                except Exception as ex:
                    print "Unable to open file " + line_ar[1] + " because: " + str(ex)
                    sys.exit(1)
                content = f.read()
                f.close()
                filename = os.path.join(*(line_ar[2].split("/")))
                try:
                    f = open(filename, "w")
                except Exception as ex:
                    print "Unable to open file " + line_ar[2] + " because: " + str(ex)
                    sys.exit(1)
                f.write(content)
                f.close()
            elif line_ar[0] == "open":
                response = br.open(line_ar[1])
                br._factory.is_html = True
                curr_test_passed = check_syntax(response.get_data(), parser, lineno) and curr_test_passed
            elif line_ar[0] == "form":
                br.select_form(line_ar[1])
                inputs = {}
                for opt in line_ar[3:]:
                    if opt[0] != '"':
                        multiple = (opt[-1] == "+")
                        opt = opt[:-1]
                        inputs[opt] = [multiple, []]
                        last = opt
                    else:
                        inputs[last][1] += [opt[1:-1]]
                for name in inputs:
                    br[name] = inputs[name][1][0] if not(inputs[name][0]) else inputs[name][1]
                if line_ar[2] != "":
                    response = br.submit(line_ar[2])
                else:
                    response = br.submit()
                br._factory.is_html = True
                curr_test_passed = check_syntax(response.get_data(), parser, lineno) and curr_test_passed
            elif line_ar[0] == "link":
                try:
                    req = br.click_link(name=line_ar[1])
                    response = br.open(req)
                    br._factory.is_html = True
                    curr_test_passed = check_syntax(response.get_data(), parser, lineno) and curr_test_passed
                except Exception as ex:
                    print "Cannot find link %s in HTML file obtained in line %d, saved in file errors/error.%d.html" %(line_ar[1], lineno, lineno)
                    filename = "errors/error.%d.html" %lineno
                    f = open(filename, "w")
                    f.write(response.get_data())
                    f.close()
                    curr_test_passed = False
            elif line_ar[0] == "save":
                filename = os.path.join(*(line_ar[1].split("/")))
                try:
                    f = open(filename, "w")
                except Exception as ex:
                    print "Unable to open file " + line_ar[1] + " because: " + str(ex)
                    sys.exit(1)
                f.write(response.get_data())
                f.close()
            elif line_ar[0] == "clear_cookies":
                cj.clear()
            elif line_ar[0] == "check_present":
                indexes = [response.get_data().find(item) for item in line_ar[1:]]
                indexes_sort = indexes[:]
                indexes_sort.sort()
                if indexes[:] != indexes_sort[:] or indexes_sort[0] == -1:
                    print "HTML file obtained in line %d does not pass check_present, saved in file errors/error.%d.html" %(lineno, lineno)
                    filename = "errors/error.%d.html" %lineno
                    f = open(filename, "w")
                    f.write("<!--Strings searched, should be present in ascending order: %s-->\n" %str(line_ar[1:]))
                    f.write("<!--Indexes: %s-->\n" %str(indexes))
                    f.write(response.get_data())
                    f.close()
                    curr_test_passed = False
            elif line_ar[0] == "check_not_present":
                indexes = [response.get_data().find(item) for item in line_ar[1:]]
                if set(indexes) != set([-1]):
                    print "HTML file obtained in line %d does not pass check_not_present, saved in file errors/error.%d.html" %(lineno, lineno)
                    filename = "errors/error.%d.html" %lineno
                    f = open(filename, "w")
                    f.write("<!--Strings searched should be NOT present: %s-->\n" %str(line_ar[1:]))
                    f.write("<!--Indexes: %s-->\n" %str(indexes))
                    f.write(response.get_data())
                    f.close()
                    curr_test_passed = False
            elif line_ar[0] == "block_cookies":
                # pol = cookielib.DefaultCookiePolicy()
                # pol.set_blocked_domains([line_ar[1]])
                # cj.set_policy(pol)
                cj.set_policy(cookielib.DefaultCookiePolicy(rfc2965=True, strict_ns_domain=cookielib.DefaultCookiePolicy.DomainStrict, allowed_domains=["out.com"]))
            elif line_ar[0] == "allow_cookies":
                # pol = cookielib.DefaultCookiePolicy()
                # pol.set_blocked_domains([])
                # cj.set_policy(pol)
                cj.set_policy(cookielib.DefaultCookiePolicy(rfc2965=True, strict_ns_domain=cookielib.DefaultCookiePolicy.DomainStrict, allowed_domains=None))
            elif line_ar[0] == "reload":
                br_man = Browser()
                cj_man = cookielib.CookieJar()
                br_man.set_cookiejar(cj_man)
                br_man.add_password('http://localhost:8080/manager/html/', login_manager, password_manager)
                response_man = br_man.open("http://localhost:8080/manager/html/")
                lines_man = response_man.get_data().splitlines()
                count_form = -1
                for line_man in lines_man:
                    if line_man.find("<form") != -1:
                        count_form += 1
                    if line_man.find("Reload") != -1 and line_man.find("ecommerce%%5F2016%%5F%s" %matricola) != -1:
                        break
                br_man.select_form(nr = count_form)
                br_man.submit()
            elif line_ar[0] == "init":
                login_manager = line_ar[1]
                password_manager = line_ar[2]
                matricola = line_ar[3]
            elif line_ar[0] == "xml_modify":
                xml_modify(line_ar[1], line_ar[2], line_ar[3])
            elif line_ar[0] == "check_image_present":
                src_image = get_image(response.get_data(), line_ar)
                if src_image == None:
                    print "HTML file obtained in line %d does not pass %s, saved in file errors/error.%d.html" %(lineno, line_ar[0], lineno)
                    filename = "errors/error.%d.html" %lineno
                    f = open(filename, "w")
                    f.write("<!--Searched image %s, not found-->\n" %str(line_ar[1]))
                    f.write(response.get_data())
                    f.close()
                    curr_test_passed = False
                else:
                    try:
                        if src_image[0] == '/':
                            f = br.retrieve(line_ar[2] + "/" + src_image)[0]
                        else:
                            f = br.retrieve(src_image)[0]
                    except Exception as ex:
                        print "HTML file obtained in line %d does not pass %s, saved in file errors/error.%d.html" %(lineno, line_ar[0], lineno)
                        filename = "errors/error.%d.html" %lineno
                        f = open(filename, "w")
                        f.write("<!--Searched image %s, i.e. %s; should be downloadable, but it is not-->\n" %(str(line_ar[1]), src_image))
                        f.write(response.get_data())
                        f.close()
                        curr_test_passed = False
            elif line_ar[0] == "check_image_not_present":
                src_image = get_image(response.get_data(), line_ar)
                if src_image == None:
                    print "HTML file obtained in line %d does not pass %s, saved in file errors/error.%d.html" %(lineno, line_ar[0], lineno)
                    filename = "errors/error.%d.html" %lineno
                    f = open(filename, "w")
                    f.write("<!--Searched image %s, not found-->\n" %str(line_ar[1]))
                    f.write(response.get_data())
                    f.close()
                    curr_test_passed = False
                else:
                    try:
                        if src_image[0] == '/':
                            ftmp = br.retrieve(line_ar[2] + src_image)[0]
                        else:
                            ftmp = br.retrieve(src_image)[0]
                        print "HTML file obtained in line %d does not pass %s, saved in file errors/error.%d.html and errors/error_additional.%d.html" %(lineno, line_ar[0], lineno, lineno)
                        filename = "errors/error.%d.html" %lineno
                        f = open(filename, "w")
                        f.write("<!--Searched image %s, i.e. %s; should be non-downloadable, but it seems something is downloaded (download result is in file errors/error_additional.%d.html)-->\n" %(str(line_ar[1]), src_image, lineno))
                        f.write(response.get_data())
                        f.close()
                        filename = "errors/error_additional.%d.html" %lineno
                        f = open(filename, "w")
                        g = open(ftmp, "r")
                        f.write("%s" %(g.read()))
                        f.close()
                        g.close()
                        curr_test_passed = False
                    except Exception as ex:
                        pass
            elif line_ar[0] == "delete_all_db":
                for table in ["account", "currency", "item", "category", "categ_item"]:
                    try:
                        exec_and_retrieve(db, "drop table %s cascade" %table, False)
                    except Exception as ex:
                        raise Exception("Unable to delete table " + table + " because of " + str(ex))
                d = open(db_file, "r")
                db_ddl = d.readlines()
                d.close()
                inside = False
                sql = ""
                for line_ddl in db_ddl:
                    if not(inside) and line_ddl.find("create table") == 0:
                        inside = True
                        sql += line_ddl.replace("\n", "")
                    elif inside and line_ddl == ");\n":
                        inside = False
                        sql += ")"
                        try:
                            exec_and_retrieve(db, sql, False)
                        except Exception as ex:
                            raise Exception("Unable to perform DDL query " + sql + " because of " + str(ex))
                        sql = ""
                    elif inside:
                        sql += line_ddl.replace("\n", "")
            elif line_ar[0] == "load_database_from_xml":
                try:
                    from_xml_to_sql(db, line_ar[1])
                except Exception as ex:
                    raise Exception("Unable to load DB from XML file " + line_ar[1] + " because of " + str(ex))
            elif line_ar[0] == "check_if_equal":
                htmls = {}
                for i in [1, 2]:
                    response = br.open(line_ar[i])
                    br._factory.is_html = True
                    htmls[i] = response.get_data()
                if htmls[1] != htmls[2]:
                    print "HTML files %s and %s obtained in line %d does not pass check_if_equal" %(line_ar[1], line_ar[2], lineno)
                    curr_test_passed = False
            elif line_ar[0] == "exit":
                raise Exception("Exited")
    except Exception as ex:
        print "\n\nSomething went wrong at line %d of %s (line %d of main.py): %s\nLast available page is errors/error.%d.html" %(lineno, test_file, sys.exc_info()[-1].tb_lineno, str(ex), lineno)
        f = open("errors/error.%d.html" %lineno, "w")
        f.write(response.get_data())
        f.close()
        curr_test_passed = False
print "\n\nPROJECT EVALUATION: " + str(points) + "/" + str(total_points) + "\n"
db.close()
conn.close()
